@@ -254,6 +254,7 @@ struct ckmap mechanism_names[] = {
     MECH_ENTRY(CKM_SHA1_RSA_X9_31),
     MECH_ENTRY(CKM_RSA_PKCS_PSS),
     MECH_ENTRY(CKM_SHA1_RSA_PKCS_PSS),
+    MECH_ENTRY(CKM_ML_KEM_KEY_PAIR_GEN),
     MECH_ENTRY(CKM_DSA_KEY_PAIR_GEN),
     MECH_ENTRY(CKM_DSA),
     MECH_ENTRY(CKM_DSA_SHA1),
@@ -261,16 +262,43 @@ struct ckmap mechanism_names[] = {
     MECH_ENTRY(CKM_DSA_SHA256),
     MECH_ENTRY(CKM_DSA_SHA384),
     MECH_ENTRY(CKM_DSA_SHA512),
+    MECH_ENTRY(CKM_ML_KEM),
     MECH_ENTRY(CKM_DSA_SHA3_224),
     MECH_ENTRY(CKM_DSA_SHA3_256),
     MECH_ENTRY(CKM_DSA_SHA3_384),
     MECH_ENTRY(CKM_DSA_SHA3_512),
+    MECH_ENTRY(CKM_ML_DSA_KEY_PAIR_GEN),
+    MECH_ENTRY(CKM_ML_DSA),
+    MECH_ENTRY(CKM_HASH_ML_DSA),
     MECH_ENTRY(CKM_DH_PKCS_KEY_PAIR_GEN),
     MECH_ENTRY(CKM_DH_PKCS_DERIVE),
+    MECH_ENTRY(CKM_HASH_ML_DSA_SHA224),
+    MECH_ENTRY(CKM_HASH_ML_DSA_SHA256),
+    MECH_ENTRY(CKM_HASH_ML_DSA_SHA384),
+    MECH_ENTRY(CKM_HASH_ML_DSA_SHA512),
+    MECH_ENTRY(CKM_HASH_ML_DSA_SHA3_224),
+    MECH_ENTRY(CKM_HASH_ML_DSA_SHA3_256),
+    MECH_ENTRY(CKM_HASH_ML_DSA_SHA3_384),
+    MECH_ENTRY(CKM_HASH_ML_DSA_SHA3_512),
+    MECH_ENTRY(CKM_HASH_ML_DSA_SHAKE128),
+    MECH_ENTRY(CKM_HASH_ML_DSA_SHAKE256),
+    MECH_ENTRY(CKM_SLH_DSA_KEY_PAIR_GEN),
+    MECH_ENTRY(CKM_SLH_DSA),
     MECH_ENTRY(CKM_X9_42_DH_KEY_PAIR_GEN),
     MECH_ENTRY(CKM_X9_42_DH_DERIVE),
     MECH_ENTRY(CKM_X9_42_DH_HYBRID_DERIVE),
     MECH_ENTRY(CKM_X9_42_MQV_DERIVE),
+    MECH_ENTRY(CKM_HASH_SLH_DSA),
+    MECH_ENTRY(CKM_HASH_SLH_DSA_SHA224),
+    MECH_ENTRY(CKM_HASH_SLH_DSA_SHA256),
+    MECH_ENTRY(CKM_HASH_SLH_DSA_SHA384),
+    MECH_ENTRY(CKM_HASH_SLH_DSA_SHA512),
+    MECH_ENTRY(CKM_HASH_SLH_DSA_SHA3_224),
+    MECH_ENTRY(CKM_HASH_SLH_DSA_SHA3_256),
+    MECH_ENTRY(CKM_HASH_SLH_DSA_SHA3_384),
+    MECH_ENTRY(CKM_HASH_SLH_DSA_SHA3_512),
+    MECH_ENTRY(CKM_HASH_SLH_DSA_SHAKE128),
+    MECH_ENTRY(CKM_HASH_SLH_DSA_SHAKE256),
     MECH_ENTRY(CKM_SHA256_RSA_PKCS),
     MECH_ENTRY(CKM_SHA384_RSA_PKCS),
     MECH_ENTRY(CKM_SHA512_RSA_PKCS),
@@ -291,6 +319,8 @@ struct ckmap mechanism_names[] = {
     MECH_ENTRY(CKM_SHA512_T_HMAC),
     MECH_ENTRY(CKM_SHA512_T_HMAC_GENERAL),
     MECH_ENTRY(CKM_SHA512_T_KEY_DERIVATION),
+    MECH_ENTRY(CKM_TLS12_EXTENDED_MASTER_KEY_DERIVE),
+    MECH_ENTRY(CKM_TLS12_EXTENDED_MASTER_KEY_DERIVE_DH),
     MECH_ENTRY(CKM_SHA3_256_RSA_PKCS),
     MECH_ENTRY(CKM_SHA3_384_RSA_PKCS),
     MECH_ENTRY(CKM_SHA3_512_RSA_PKCS),
@@ -461,6 +491,8 @@ struct ckmap mechanism_names[] = {
     MECH_ENTRY(CKM_WTLS_PRF),
     MECH_ENTRY(CKM_WTLS_SERVER_KEY_AND_MAC_DERIVE),
     MECH_ENTRY(CKM_WTLS_CLIENT_KEY_AND_MAC_DERIVE),
+    MECH_ENTRY(CKM_TLS10_MAC_SERVER),
+    MECH_ENTRY(CKM_TLS10_MAC_CLIENT),
     MECH_ENTRY(CKM_TLS12_MAC),
     MECH_ENTRY(CKM_TLS12_KDF),
     MECH_ENTRY(CKM_TLS12_MASTER_KEY_DERIVE),
@@ -601,6 +633,7 @@ struct ckmap mechanism_names[] = {
     MECH_ENTRY(CKM_AES_KEY_WRAP),
     MECH_ENTRY(CKM_AES_KEY_WRAP_PAD),
     MECH_ENTRY(CKM_AES_KEY_WRAP_KWP),
+    MECH_ENTRY(CKM_AES_KEY_WRAP_PKCS7),
     MECH_ENTRY(CKM_RSA_PKCS_TPM_1_1),
     MECH_ENTRY(CKM_RSA_PKCS_OAEP_TPM_1_1),
     MECH_ENTRY(CKM_SHA_1_KEY_GEN),
@@ -656,6 +689,19 @@ struct ckmap mechanism_names[] = {
     MECH_ENTRY(CKM_SP800_108_COUNTER_KDF),
     MECH_ENTRY(CKM_SP800_108_FEEDBACK_KDF),
     MECH_ENTRY(CKM_SP800_108_DOUBLE_PIPELINE_KDF),
+    MECH_ENTRY(CKM_IKE2_PRF_PLUS_DERIVE),
+    MECH_ENTRY(CKM_IKE_PRF_DERIVE),
+    MECH_ENTRY(CKM_IKE1_PRF_DERIVE),
+    MECH_ENTRY(CKM_IKE1_EXTENDED_DERIVE),
+    MECH_ENTRY(CKM_HSS_KEY_PAIR_GEN),
+    MECH_ENTRY(CKM_HSS),
+    MECH_ENTRY(CKM_XMSS_KEY_PAIR_GEN),
+    MECH_ENTRY(CKM_XMSSMT_KEY_PAIR_GEN),
+    MECH_ENTRY(CKM_XMSS),
+    MECH_ENTRY(CKM_XMSSMT),
+    MECH_ENTRY(CKM_ECDH_X_AES_KEY_WRAP),
+    MECH_ENTRY(CKM_ECDH_COF_AES_KEY_WRAP),
+    MECH_ENTRY(CKM_PUB_KEY_FROM_PRIV_KEY),
     { 0, NULL },
 };
 
@@ -686,6 +732,9 @@ struct ckmap mechanism_flags[] = {
     MECH_ENTRY(CKF_EC_UNCOMPRESS),
     MECH_ENTRY(CKF_EC_COMPRESS),
     MECH_ENTRY(CKF_EC_CURVENAME),
+    MECH_ENTRY(CKF_ENCAPSULATE),
+    MECH_ENTRY(CKF_DECAPSULATE),
+    MECH_ENTRY(CKF_EXTENSION),
     { 0, NULL },
 };
 
@@ -709,6 +758,8 @@ struct ckmap token_flags[] = {
     MECH_ENTRY(CKF_SO_PIN_LOCKED),
     MECH_ENTRY(CKF_SO_PIN_TO_BE_CHANGED),
     MECH_ENTRY(CKF_ERROR_STATE),
+    MECH_ENTRY(CKF_SEED_RANDOM_REQUIRED),
+    MECH_ENTRY(CKF_ASYNC_SESSION_SUPPORTED),
     { 0, NULL },
 };
 
@@ -725,6 +776,8 @@ struct ckmap profile_ids[] = {
     MECH_ENTRY(CKP_EXTENDED_PROVIDER),
     MECH_ENTRY(CKP_AUTHENTICATION_TOKEN),
     MECH_ENTRY(CKP_PUBLIC_CERTIFICATES_TOKEN),
+    MECH_ENTRY(CKP_COMPLETE_PROVIDER),
+    MECH_ENTRY(CKP_HKDF_TLS_TOKEN),
     MECH_ENTRY(CKP_VENDOR_DEFINED),
     { 0, NULL },
 };
