@@ -2,6 +2,7 @@
    SPDX-License-Identifier: Apache-2.0 */
 
 #include "provider.h"
+#include <openssl/opensslv.h>
 #include <string.h>
 
 /* General Digest Mapping functions */
@@ -156,7 +157,7 @@ static void *p11prov_digest_dupctx(void *ctx)
     CK_SESSION_HANDLE sess = CK_INVALID_HANDLE;
     CK_BYTE_PTR state = NULL;
     CK_ULONG state_len;
-    CK_RV ret;
+    CK_RV ret = CKR_OK;
 
     P11PROV_debug("digest dupctx, ctx=%p", ctx);
 
@@ -178,27 +179,25 @@ static void *p11prov_digest_dupctx(void *ctx)
     /* This is not really funny. OpenSSL by default assumes contexts with
      * operations in flight can be easily duplicated, with all the
      * cryptographic status and then both context can keep going
-     * independently. We'll try to save/restore state here, but on failure
+     * independently. We'll try to save/restore state here. for OpenSSL versions
+     * greater than 3.2 this will be a hard failure, while for older OpenSSL
+     * versions, in order to maintain compatibility with previous behavior
      * we just 'move' the the session to the new context and hope there is no
      * need for the old context which will have no session and just return
      * errors if an update is attempted. */
 
     sess = p11prov_session_handle(dctx->session);
 
-    /* move old session to new context, we swap because often openssl continues
-     * on the duplicated context by default */
-    newctx->session = dctx->session;
-    dctx->session = NULL;
-
     /* NOTE: most tokens will probably return errors trying to do this on digest
-     * sessions. If GetOperationState fails we don't even try to duplicate the
-     * context. */
+     * sessions */
     ret = p11prov_GetOperationState(dctx->provctx, sess, NULL_PTR, &state_len);
     if (ret != CKR_OK) {
         goto done;
     }
+
     state = OPENSSL_malloc(state_len);
     if (state == NULL) {
+        ret = CKR_HOST_MEMORY;
         goto done;
     }
 
@@ -207,23 +206,34 @@ static void *p11prov_digest_dupctx(void *ctx)
         goto done;
     }
 
-    ret =
-        p11prov_get_session(dctx->provctx, &slotid, NULL, NULL, dctx->mechtype,
-                            NULL, NULL, false, false, &dctx->session);
+    ret = p11prov_get_session(newctx->provctx, &slotid, NULL, NULL,
+                              newctx->mechtype, NULL, NULL, false, false,
+                              &newctx->session);
     if (ret != CKR_OK) {
         P11PROV_raise(dctx->provctx, ret, "Failed to open new session");
         goto done;
     }
-    sess = p11prov_session_handle(dctx->session);
+    sess = p11prov_session_handle(newctx->session);
 
-    ret = p11prov_SetOperationState(dctx->provctx, sess, state, state_len,
+    ret = p11prov_SetOperationState(newctx->provctx, sess, state, state_len,
                                     CK_INVALID_HANDLE, CK_INVALID_HANDLE);
-    if (ret != CKR_OK) {
-        p11prov_return_session(dctx->session);
-        dctx->session = NULL;
-    }
 
 done:
+    if (ret != CKR_OK) {
+#if OPENSSL_VERSION_NUMBER >= 0x30200000L
+        /* For OpenSSL 3.2+ we can fail duplication */
+        p11prov_digest_freectx(newctx);
+        newctx = NULL;
+#else
+        /* For OpenSSL 3.0.x, we must 'move' the context on failure */
+        p11prov_return_session(newctx->session);
+
+        /* move old session to new context, we swap because normally openssl
+         * continues on the duplicated context */
+        newctx->session = dctx->session;
+        dctx->session = NULL;
+#endif
+    }
     OPENSSL_free(state);
     return newctx;
 }
