@@ -1,4 +1,5 @@
 /* Copyright (C) 2022 Simo Sorce <simo@redhat.com>
+   Copyright 2025 NXP
    SPDX-License-Identifier: Apache-2.0 */
 
 #include "provider.h"
@@ -53,11 +54,16 @@ struct p11prov_ctx {
     OSSL_ALGORITHM *op_exchange;
     OSSL_ALGORITHM *op_signature;
     OSSL_ALGORITHM *op_asym_cipher;
+    OSSL_ALGORITHM *op_kem;
 
     OSSL_ALGORITHM *op_encoder;
     OSSL_ALGORITHM *op_decoder;
     OSSL_ALGORITHM *op_keymgmt;
     OSSL_ALGORITHM *op_store;
+#if SKEY_SUPPORT == 1
+    OSSL_ALGORITHM *op_cipher;
+    OSSL_ALGORITHM *op_skeymgmt;
+#endif
 
     pthread_rwlock_t quirk_lock;
     struct quirk *quirks;
@@ -540,10 +546,15 @@ static void p11prov_ctx_free(P11PROV_CTX *ctx)
     OPENSSL_free(ctx->op_exchange);
     OPENSSL_free(ctx->op_signature);
     OPENSSL_free(ctx->op_asym_cipher);
+    OPENSSL_free(ctx->op_kem);
     OPENSSL_free(ctx->op_encoder);
     OPENSSL_free(ctx->op_decoder);
     OPENSSL_free(ctx->op_keymgmt);
     OPENSSL_free(ctx->op_store);
+#if SKEY_SUPPORT == 1
+    OPENSSL_free(ctx->op_cipher);
+    OPENSSL_free(ctx->op_skeymgmt);
+#endif
 
     OSSL_LIB_CTX_free(ctx->libctx);
     ctx->libctx = NULL;
@@ -753,16 +764,14 @@ static int p11prov_get_params(void *provctx, OSSL_PARAM params[])
     }
     p = OSSL_PARAM_locate(params, OSSL_PROV_PARAM_VERSION);
     if (p != NULL) {
-        /* temporarily return the OpenSSL build version */
-        ret = OSSL_PARAM_set_utf8_ptr(p, OPENSSL_VERSION_STR);
+        ret = OSSL_PARAM_set_utf8_ptr(p, P11PROV_VERSION);
         if (ret == 0) {
             return RET_OSSL_ERR;
         }
     }
     p = OSSL_PARAM_locate(params, OSSL_PROV_PARAM_BUILDINFO);
     if (p != NULL) {
-        /* temporarily return the OpenSSL build version */
-        ret = OSSL_PARAM_set_utf8_ptr(p, OPENSSL_FULL_VERSION_STR);
+        ret = OSSL_PARAM_set_utf8_ptr(p, P11PROV_BUILDINFO);
         if (ret == 0) {
             return RET_OSSL_ERR;
         }
@@ -849,6 +858,16 @@ static CK_RV alg_set_op(OSSL_ALGORITHM **op, int idx, OSSL_ALGORITHM *alg)
         CKM_ECDSA_SHA384, CKM_ECDSA_SHA512, CKM_ECDSA_SHA3_224, \
         CKM_ECDSA_SHA3_256, CKM_ECDSA_SHA3_384, CKM_ECDSA_SHA3_512
 
+#define PQC_MECHS \
+    CKM_ML_DSA, CKM_ML_DSA_KEY_PAIR_GEN, CKM_ML_KEM, CKM_ML_KEM_KEY_PAIR_GEN
+
+#if SKEY_SUPPORT == 1
+#define AES_MECHS \
+    CKM_AES_ECB, CKM_AES_CBC, CKM_AES_CBC_PAD, CKM_AES_CTR, CKM_AES_CTS, \
+        CKM_AES_OFB, CKM_AES_CFB8, CKM_AES_CFB128, CKM_AES_CFB1, CKM_AES_GCM, \
+        CKM_CHACHA20_POLY1305
+#endif
+
 static void alg_rm_mechs(CK_ULONG *checklist, CK_ULONG *rmlist, int *clsize,
                          int rmsize)
 {
@@ -895,14 +914,22 @@ static CK_RV operations_init(P11PROV_CTX *ctx)
                              RSAPSS_SIG_MECHS,
                              RSA_ENC_MECHS,
                              CKM_EC_KEY_PAIR_GEN,
+                             CKM_EC_EDWARDS_KEY_PAIR_GEN,
                              ECDSA_SIG_MECHS,
                              CKM_ECDH1_DERIVE,
                              CKM_ECDH1_COFACTOR_DERIVE,
+                             CKM_EC_MONTGOMERY_KEY_PAIR_GEN,
                              CKM_HKDF_DERIVE,
                              DIGEST_MECHS,
-                             CKM_EDDSA };
+                             CKM_EDDSA,
+                             PQC_MECHS,
+#if SKEY_SUPPORT == 1
+                             AES_MECHS
+#endif
+    };
     bool add_rsasig = false;
     bool add_rsaenc = false;
+    bool add_ecdsasig = false;
     int cl_size = sizeof(checklist) / sizeof(CK_ULONG);
     int digest_idx = 0;
     int kdf_idx = 0;
@@ -910,6 +937,10 @@ static CK_RV operations_init(P11PROV_CTX *ctx)
     int exchange_idx = 0;
     int signature_idx = 0;
     int asym_cipher_idx = 0;
+    int kem_idx = 0;
+#if SKEY_SUPPORT == 1
+    int cipher_idx = 0;
+#endif
     int slot_idx = 0;
     const char *prop = get_default_properties(ctx);
     CK_RV ret;
@@ -947,9 +978,65 @@ static CK_RV operations_init(P11PROV_CTX *ctx)
             case CKM_RSA_PKCS:
                 add_rsasig = true;
                 add_rsaenc = true;
-                UNCHECK_MECHS(CKM_RSA_PKCS_KEY_PAIR_GEN, RSA_SIG_MECHS);
+                UNCHECK_MECHS(CKM_RSA_PKCS_KEY_PAIR_GEN);
                 UNCHECK_MECHS(CKM_RSA_PKCS_KEY_PAIR_GEN, RSA_ENC_MECHS);
                 break;
+#if defined(OSSL_FUNC_SIGNATURE_SIGN_MESSAGE_INIT)
+            case CKM_SHA1_RSA_PKCS:
+                ADD_ALGO_EXT(RSA_SHA1, signature, prop,
+                             p11prov_rsa_sha1_signature_functions);
+                add_rsasig = true;
+                UNCHECK_MECHS(CKM_SHA1_RSA_PKCS, CKM_RSA_PKCS_KEY_PAIR_GEN);
+                break;
+            case CKM_SHA224_RSA_PKCS:
+                ADD_ALGO_EXT(RSA_SHA224, signature, prop,
+                             p11prov_rsa_sha224_signature_functions);
+                add_rsasig = true;
+                UNCHECK_MECHS(CKM_SHA224_RSA_PKCS, CKM_RSA_PKCS_KEY_PAIR_GEN);
+                break;
+            case CKM_SHA256_RSA_PKCS:
+                ADD_ALGO_EXT(RSA_SHA256, signature, prop,
+                             p11prov_rsa_sha256_signature_functions);
+                add_rsasig = true;
+                UNCHECK_MECHS(CKM_SHA256_RSA_PKCS, CKM_RSA_PKCS_KEY_PAIR_GEN);
+                break;
+            case CKM_SHA384_RSA_PKCS:
+                ADD_ALGO_EXT(RSA_SHA384, signature, prop,
+                             p11prov_rsa_sha384_signature_functions);
+                add_rsasig = true;
+                UNCHECK_MECHS(CKM_SHA384_RSA_PKCS, CKM_RSA_PKCS_KEY_PAIR_GEN);
+                break;
+            case CKM_SHA512_RSA_PKCS:
+                ADD_ALGO_EXT(RSA_SHA512, signature, prop,
+                             p11prov_rsa_sha512_signature_functions);
+                add_rsasig = true;
+                UNCHECK_MECHS(CKM_SHA512_RSA_PKCS, CKM_RSA_PKCS_KEY_PAIR_GEN);
+                break;
+            case CKM_SHA3_224_RSA_PKCS:
+                ADD_ALGO_EXT(RSA_SHA3_224, signature, prop,
+                             p11prov_rsa_sha3_224_signature_functions);
+                add_rsasig = true;
+                UNCHECK_MECHS(CKM_SHA3_224_RSA_PKCS, CKM_RSA_PKCS_KEY_PAIR_GEN);
+                break;
+            case CKM_SHA3_256_RSA_PKCS:
+                ADD_ALGO_EXT(RSA_SHA3_256, signature, prop,
+                             p11prov_rsa_sha3_256_signature_functions);
+                add_rsasig = true;
+                UNCHECK_MECHS(CKM_SHA3_256_RSA_PKCS, CKM_RSA_PKCS_KEY_PAIR_GEN);
+                break;
+            case CKM_SHA3_384_RSA_PKCS:
+                ADD_ALGO_EXT(RSA_SHA3_384, signature, prop,
+                             p11prov_rsa_sha3_384_signature_functions);
+                add_rsasig = true;
+                UNCHECK_MECHS(CKM_SHA3_384_RSA_PKCS, CKM_RSA_PKCS_KEY_PAIR_GEN);
+                break;
+            case CKM_SHA3_512_RSA_PKCS:
+                ADD_ALGO_EXT(RSA_SHA3_512, signature, prop,
+                             p11prov_rsa_sha3_512_signature_functions);
+                add_rsasig = true;
+                UNCHECK_MECHS(CKM_SHA3_512_RSA_PKCS, CKM_RSA_PKCS_KEY_PAIR_GEN);
+                break;
+#else
             case CKM_SHA1_RSA_PKCS:
             case CKM_SHA224_RSA_PKCS:
             case CKM_SHA256_RSA_PKCS:
@@ -962,6 +1049,7 @@ static CK_RV operations_init(P11PROV_CTX *ctx)
                 add_rsasig = true;
                 UNCHECK_MECHS(CKM_RSA_PKCS_KEY_PAIR_GEN, RSA_SIG_MECHS);
                 break;
+#endif
             case CKM_RSA_PKCS_PSS:
             case CKM_SHA1_RSA_PKCS_PSS:
             case CKM_SHA224_RSA_PKCS_PSS:
@@ -985,6 +1073,65 @@ static CK_RV operations_init(P11PROV_CTX *ctx)
                 UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN);
                 break;
             case CKM_ECDSA:
+                add_ecdsasig = true;
+                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, CKM_ECDSA);
+                break;
+#if defined(OSSL_FUNC_SIGNATURE_SIGN_MESSAGE_INIT)
+            case CKM_ECDSA_SHA1:
+                ADD_ALGO_EXT(ECDSA_SHA1, signature, prop,
+                             p11prov_ecdsa_sha1_signature_functions);
+                add_ecdsasig = true;
+                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, CKM_ECDSA_SHA1);
+                break;
+            case CKM_ECDSA_SHA224:
+                ADD_ALGO_EXT(ECDSA_SHA224, signature, prop,
+                             p11prov_ecdsa_sha224_signature_functions);
+                add_ecdsasig = true;
+                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, CKM_ECDSA_SHA224);
+                break;
+            case CKM_ECDSA_SHA256:
+                ADD_ALGO_EXT(ECDSA_SHA256, signature, prop,
+                             p11prov_ecdsa_sha256_signature_functions);
+                add_ecdsasig = true;
+                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, CKM_ECDSA_SHA256);
+                break;
+            case CKM_ECDSA_SHA384:
+                ADD_ALGO_EXT(ECDSA_SHA384, signature, prop,
+                             p11prov_ecdsa_sha384_signature_functions);
+                add_ecdsasig = true;
+                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, CKM_ECDSA_SHA384);
+                break;
+            case CKM_ECDSA_SHA512:
+                ADD_ALGO_EXT(ECDSA_SHA512, signature, prop,
+                             p11prov_ecdsa_sha512_signature_functions);
+                add_ecdsasig = true;
+                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, CKM_ECDSA_SHA512);
+                break;
+            case CKM_ECDSA_SHA3_224:
+                ADD_ALGO_EXT(ECDSA_SHA3_224, signature, prop,
+                             p11prov_ecdsa_sha3_224_signature_functions);
+                add_ecdsasig = true;
+                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, CKM_ECDSA_SHA3_224);
+                break;
+            case CKM_ECDSA_SHA3_256:
+                ADD_ALGO_EXT(ECDSA_SHA3_256, signature, prop,
+                             p11prov_ecdsa_sha3_256_signature_functions);
+                add_ecdsasig = true;
+                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, CKM_ECDSA_SHA3_256);
+                break;
+            case CKM_ECDSA_SHA3_384:
+                ADD_ALGO_EXT(ECDSA_SHA3_384, signature, prop,
+                             p11prov_ecdsa_sha3_384_signature_functions);
+                add_ecdsasig = true;
+                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, CKM_ECDSA_SHA3_384);
+                break;
+            case CKM_ECDSA_SHA3_512:
+                ADD_ALGO_EXT(ECDSA_SHA3_512, signature, prop,
+                             p11prov_ecdsa_sha3_512_signature_functions);
+                add_ecdsasig = true;
+                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, CKM_ECDSA_SHA3_512);
+                break;
+#else
             case CKM_ECDSA_SHA1:
             case CKM_ECDSA_SHA224:
             case CKM_ECDSA_SHA256:
@@ -994,9 +1141,14 @@ static CK_RV operations_init(P11PROV_CTX *ctx)
             case CKM_ECDSA_SHA3_256:
             case CKM_ECDSA_SHA3_384:
             case CKM_ECDSA_SHA3_512:
-                ADD_ALGO(ECDSA, ecdsa, signature, prop);
-                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, ECDSA_SIG_MECHS);
+                add_ecdsasig = true;
+                UNCHECK_MECHS(CKM_EC_KEY_PAIR_GEN, CKM_ECDSA_SHA1,
+                              CKM_ECDSA_SHA224, CKM_ECDSA_SHA256,
+                              CKM_ECDSA_SHA384, CKM_ECDSA_SHA512,
+                              CKM_ECDSA_SHA3_224, CKM_ECDSA_SHA3_256,
+                              CKM_ECDSA_SHA3_384, CKM_ECDSA_SHA3_512);
                 break;
+#endif
             case CKM_ECDH1_DERIVE:
             case CKM_ECDH1_COFACTOR_DERIVE:
                 ADD_ALGO(ECDH, ecdh, exchange, prop);
@@ -1054,15 +1206,110 @@ static CK_RV operations_init(P11PROV_CTX *ctx)
                 UNCHECK_MECHS(CKM_SHA3_512);
                 break;
             case CKM_EDDSA:
+            case CKM_EC_EDWARDS_KEY_PAIR_GEN:
                 ADD_ALGO_EXT(ED25519, signature, prop,
-                             p11prov_eddsa_signature_functions);
+                             p11prov_ed25519_signature_functions);
                 ADD_ALGO_EXT(ED448, signature, prop,
-                             p11prov_eddsa_signature_functions);
+                             p11prov_ed448_signature_functions);
                 UNCHECK_MECHS(CKM_EC_EDWARDS_KEY_PAIR_GEN, CKM_EDDSA);
+#if defined(OSSL_FUNC_SIGNATURE_SIGN_MESSAGE_INIT)
+                ADD_ALGO_EXT(ED25519ph, signature, prop,
+                             p11prov_ed25519ph_signature_functions);
+                ADD_ALGO_EXT(ED25519ctx, signature, prop,
+                             p11prov_ed25519ctx_signature_functions);
+                ADD_ALGO_EXT(ED448ph, signature, prop,
+                             p11prov_ed448ph_signature_functions);
+#endif
                 break;
+            case CKM_EC_MONTGOMERY_KEY_PAIR_GEN:
+                ADD_ALGO(X25519, x25519, exchange, prop);
+                ADD_ALGO(X448, x448, exchange, prop);
+                break;
+            case CKM_ML_DSA:
+            case CKM_ML_DSA_KEY_PAIR_GEN:
+                ADD_ALGO_EXT(ML_DSA_44, signature, prop,
+                             p11prov_mldsa_44_signature_functions);
+                ADD_ALGO_EXT(ML_DSA_65, signature, prop,
+                             p11prov_mldsa_65_signature_functions);
+                ADD_ALGO_EXT(ML_DSA_87, signature, prop,
+                             p11prov_mldsa_87_signature_functions);
+                UNCHECK_MECHS(CKM_ML_DSA_KEY_PAIR_GEN, CKM_ML_DSA);
+                break;
+            case CKM_ML_KEM:
+            case CKM_ML_KEM_KEY_PAIR_GEN:
+                ADD_ALGO_EXT(ML_KEM_512, kem, prop,
+                             p11prov_mlkem_kem_functions);
+                ADD_ALGO_EXT(ML_KEM_768, kem, prop,
+                             p11prov_mlkem_kem_functions);
+                ADD_ALGO_EXT(ML_KEM_1024, kem, prop,
+                             p11prov_mlkem_kem_functions);
+                UNCHECK_MECHS(CKM_ML_KEM_KEY_PAIR_GEN, CKM_ML_KEM);
+                break;
+#if SKEY_SUPPORT == 1
+            case CKM_AES_ECB:
+                ADD_ALGO(AES_256_ECB, aes256ecb, cipher, prop);
+                ADD_ALGO(AES_192_ECB, aes192ecb, cipher, prop);
+                ADD_ALGO(AES_128_ECB, aes128ecb, cipher, prop);
+                UNCHECK_MECHS(CKM_AES_ECB);
+                break;
+            case CKM_AES_CBC:
+            case CKM_AES_CBC_PAD:
+                ADD_ALGO(AES_256_CBC, aes256cbc, cipher, prop);
+                ADD_ALGO(AES_192_CBC, aes192cbc, cipher, prop);
+                ADD_ALGO(AES_128_CBC, aes128cbc, cipher, prop);
+                UNCHECK_MECHS(CKM_AES_CBC);
+                UNCHECK_MECHS(CKM_AES_CBC_PAD);
+                break;
+            case CKM_AES_OFB:
+                ADD_ALGO(AES_256_OFB, aes256ofb, cipher, prop);
+                ADD_ALGO(AES_192_OFB, aes192ofb, cipher, prop);
+                ADD_ALGO(AES_128_OFB, aes128ofb, cipher, prop);
+                UNCHECK_MECHS(CKM_AES_OFB);
+                break;
+            case CKM_AES_CFB128:
+                ADD_ALGO(AES_256_CFB, aes256cfb, cipher, prop);
+                ADD_ALGO(AES_192_CFB, aes192cfb, cipher, prop);
+                ADD_ALGO(AES_128_CFB, aes128cfb, cipher, prop);
+                UNCHECK_MECHS(CKM_AES_CFB128);
+                break;
+            case CKM_AES_CFB1:
+                ADD_ALGO(AES_256_CFB1, aes256cfb1, cipher, prop);
+                ADD_ALGO(AES_192_CFB1, aes192cfb1, cipher, prop);
+                ADD_ALGO(AES_128_CFB1, aes128cfb1, cipher, prop);
+                UNCHECK_MECHS(CKM_AES_CFB1);
+                break;
+            case CKM_AES_CFB8:
+                ADD_ALGO(AES_256_CFB8, aes256cfb8, cipher, prop);
+                ADD_ALGO(AES_192_CFB8, aes192cfb8, cipher, prop);
+                ADD_ALGO(AES_128_CFB8, aes128cfb8, cipher, prop);
+                UNCHECK_MECHS(CKM_AES_CFB8);
+                break;
+            case CKM_AES_CTR:
+                ADD_ALGO(AES_256_CTR, aes256ctr, cipher, prop);
+                ADD_ALGO(AES_192_CTR, aes192ctr, cipher, prop);
+                ADD_ALGO(AES_128_CTR, aes128ctr, cipher, prop);
+                UNCHECK_MECHS(CKM_AES_CTR);
+                break;
+            case CKM_AES_CTS:
+                ADD_ALGO(AES_256_CTS, aes256cts, cipher, prop);
+                ADD_ALGO(AES_192_CTS, aes192cts, cipher, prop);
+                ADD_ALGO(AES_128_CTS, aes128cts, cipher, prop);
+                UNCHECK_MECHS(CKM_AES_CTS);
+                break;
+            case CKM_AES_GCM:
+                ADD_ALGO(AES_256_GCM, aes256gcm, cipher, prop);
+                ADD_ALGO(AES_192_GCM, aes192gcm, cipher, prop);
+                ADD_ALGO(AES_128_GCM, aes128gcm, cipher, prop);
+                UNCHECK_MECHS(CKM_AES_GCM);
+                break;
+            case CKM_CHACHA20_POLY1305:
+                ADD_ALGO(CHACHA20_POLY1305, chacha20256poly1305, cipher, prop);
+                UNCHECK_MECHS(CKM_CHACHA20_POLY1305);
+                break;
+#endif
             default:
-                P11PROV_raise(ctx, CKR_GENERAL_ERROR,
-                              "Unhandled mechianism %lu", mech);
+                P11PROV_raise(ctx, CKR_GENERAL_ERROR, "Unhandled mechanism %lu",
+                              mech);
                 break;
             }
         }
@@ -1073,6 +1320,9 @@ static CK_RV operations_init(P11PROV_CTX *ctx)
     if (add_rsasig) {
         ADD_ALGO(RSA, rsa, signature, prop);
     }
+    if (add_ecdsasig) {
+        ADD_ALGO(ECDSA, ecdsa, signature, prop);
+    }
     if (add_rsaenc) {
         ADD_ALGO(RSA, rsa, asym_cipher, prop);
     }
@@ -1082,6 +1332,10 @@ static CK_RV operations_init(P11PROV_CTX *ctx)
     TERM_ALGO(exchange);
     TERM_ALGO(signature);
     TERM_ALGO(asym_cipher);
+    TERM_ALGO(kem);
+#if SKEY_SUPPORT == 1
+    TERM_ALGO(cipher);
+#endif
 
     /* handle random */
     ret = p11prov_check_random(ctx);
@@ -1099,6 +1353,9 @@ static CK_RV static_operations_init(P11PROV_CTX *ctx)
     int decoder_idx = 0;
     int store_idx = 0;
     int keymgmt_idx = 0;
+#if SKEY_SUPPORT == 1
+    int skeymgmt_idx = 0;
+#endif
     const char *prop = get_default_properties(ctx);
 
 /* encoder */
@@ -1145,6 +1402,40 @@ static CK_RV static_operations_init(P11PROV_CTX *ctx)
                  p11prov_ec_edwards_encoder_text_functions);
     ADD_ALGO_EXT(ED448, encoder, DEFAULT_PROPERTY(",output=text"),
                  p11prov_ec_edwards_encoder_text_functions);
+    ADD_ALGO_EXT(X25519, encoder, DEFAULT_PROPERTY(",output=text"),
+                 p11prov_ec_montgomery_encoder_text_functions);
+    ADD_ALGO_EXT(X448, encoder, DEFAULT_PROPERTY(",output=text"),
+                 p11prov_ec_montgomery_encoder_text_functions);
+    ADD_ALGO_EXT(ML_DSA_44, encoder, DEFAULT_PROPERTY(",output=text"),
+                 p11prov_mldsa_encoder_text_functions);
+    ADD_ALGO_EXT(ML_DSA_44, encoder,
+                 DEFAULT_PROPERTY(",output=der,structure=SubjectPublicKeyInfo"),
+                 p11prov_mldsa_encoder_spki_der_functions);
+    ADD_ALGO_EXT(ML_DSA_65, encoder, DEFAULT_PROPERTY(",output=text"),
+                 p11prov_mldsa_encoder_text_functions);
+    ADD_ALGO_EXT(ML_DSA_65, encoder,
+                 DEFAULT_PROPERTY(",output=der,structure=SubjectPublicKeyInfo"),
+                 p11prov_mldsa_encoder_spki_der_functions);
+    ADD_ALGO_EXT(ML_DSA_87, encoder, DEFAULT_PROPERTY(",output=text"),
+                 p11prov_mldsa_encoder_text_functions);
+    ADD_ALGO_EXT(ML_DSA_87, encoder,
+                 DEFAULT_PROPERTY(",output=der,structure=SubjectPublicKeyInfo"),
+                 p11prov_mldsa_encoder_spki_der_functions);
+    ADD_ALGO_EXT(ML_KEM_512, encoder, DEFAULT_PROPERTY(",output=text"),
+                 p11prov_mlkem_encoder_text_functions);
+    ADD_ALGO_EXT(ML_KEM_512, encoder,
+                 DEFAULT_PROPERTY(",output=der,structure=SubjectPublicKeyInfo"),
+                 p11prov_mlkem_encoder_spki_der_functions);
+    ADD_ALGO_EXT(ML_KEM_768, encoder, DEFAULT_PROPERTY(",output=text"),
+                 p11prov_mlkem_encoder_text_functions);
+    ADD_ALGO_EXT(ML_KEM_768, encoder,
+                 DEFAULT_PROPERTY(",output=der,structure=SubjectPublicKeyInfo"),
+                 p11prov_mlkem_encoder_spki_der_functions);
+    ADD_ALGO_EXT(ML_KEM_1024, encoder, DEFAULT_PROPERTY(",output=text"),
+                 p11prov_mlkem_encoder_text_functions);
+    ADD_ALGO_EXT(ML_KEM_1024, encoder,
+                 DEFAULT_PROPERTY(",output=der,structure=SubjectPublicKeyInfo"),
+                 p11prov_mlkem_encoder_spki_der_functions);
     if (ctx->encode_pkey_as_pk11_uri) {
         ADD_ALGO_EXT(RSA, encoder,
                      DEFAULT_PROPERTY(",output=pem,structure=PrivateKeyInfo"),
@@ -1161,6 +1452,30 @@ static CK_RV static_operations_init(P11PROV_CTX *ctx)
         ADD_ALGO_EXT(ED448, encoder,
                      DEFAULT_PROPERTY(",output=pem,structure=PrivateKeyInfo"),
                      p11prov_ec_edwards_encoder_priv_key_info_pem_functions);
+        ADD_ALGO_EXT(X25519, encoder,
+                     DEFAULT_PROPERTY(",output=pem,structure=PrivateKeyInfo"),
+                     p11prov_ec_montgomery_encoder_priv_key_info_pem_functions);
+        ADD_ALGO_EXT(X448, encoder,
+                     DEFAULT_PROPERTY(",output=pem,structure=PrivateKeyInfo"),
+                     p11prov_ec_montgomery_encoder_priv_key_info_pem_functions);
+        ADD_ALGO_EXT(ML_DSA_44, encoder,
+                     DEFAULT_PROPERTY(",output=pem,structure=PrivateKeyInfo"),
+                     p11prov_mldsa_encoder_priv_key_info_pem_functions);
+        ADD_ALGO_EXT(ML_DSA_65, encoder,
+                     DEFAULT_PROPERTY(",output=pem,structure=PrivateKeyInfo"),
+                     p11prov_mldsa_encoder_priv_key_info_pem_functions);
+        ADD_ALGO_EXT(ML_DSA_87, encoder,
+                     DEFAULT_PROPERTY(",output=pem,structure=PrivateKeyInfo"),
+                     p11prov_mldsa_encoder_priv_key_info_pem_functions);
+        ADD_ALGO_EXT(ML_KEM_512, encoder,
+                     DEFAULT_PROPERTY(",output=pem,structure=PrivateKeyInfo"),
+                     p11prov_mlkem_encoder_priv_key_info_pem_functions);
+        ADD_ALGO_EXT(ML_KEM_768, encoder,
+                     DEFAULT_PROPERTY(",output=pem,structure=PrivateKeyInfo"),
+                     p11prov_mlkem_encoder_priv_key_info_pem_functions);
+        ADD_ALGO_EXT(ML_KEM_1024, encoder,
+                     DEFAULT_PROPERTY(",output=pem,structure=PrivateKeyInfo"),
+                     p11prov_mlkem_encoder_priv_key_info_pem_functions);
     }
 
     TERM_ALGO(encoder);
@@ -1179,6 +1494,10 @@ static CK_RV static_operations_init(P11PROV_CTX *ctx)
                  p11prov_der_decoder_p11prov_ed25519_functions);
     ADD_ALGO_EXT(ED448, decoder, DEFAULT_PROPERTY(DER_DECODER_PROP),
                  p11prov_der_decoder_p11prov_ed448_functions);
+    ADD_ALGO_EXT(X25519, decoder, DEFAULT_PROPERTY(DER_DECODER_PROP),
+                 p11prov_der_decoder_p11prov_x25519_functions);
+    ADD_ALGO_EXT(X448, decoder, DEFAULT_PROPERTY(DER_DECODER_PROP),
+                 p11prov_der_decoder_p11prov_x448_functions);
     TERM_ALGO(decoder);
 #undef DEFAULT_PROPERTY
 
@@ -1193,14 +1512,32 @@ static CK_RV static_operations_init(P11PROV_CTX *ctx)
     ADD_ALGO(HKDF, hkdf, keymgmt, prop);
     ADD_ALGO_EXT(ED25519, keymgmt, prop, p11prov_ed25519_keymgmt_functions);
     ADD_ALGO_EXT(ED448, keymgmt, prop, p11prov_ed448_keymgmt_functions);
+    ADD_ALGO_EXT(X25519, keymgmt, prop, p11prov_x25519_keymgmt_functions);
+    ADD_ALGO_EXT(X448, keymgmt, prop, p11prov_x448_keymgmt_functions);
+    ADD_ALGO_EXT(ML_DSA_44, keymgmt, prop, p11prov_mldsa44_keymgmt_functions);
+    ADD_ALGO_EXT(ML_DSA_65, keymgmt, prop, p11prov_mldsa65_keymgmt_functions);
+    ADD_ALGO_EXT(ML_DSA_87, keymgmt, prop, p11prov_mldsa87_keymgmt_functions);
+    ADD_ALGO_EXT(ML_KEM_512, keymgmt, prop, p11prov_mlkem512_keymgmt_functions);
+    ADD_ALGO_EXT(ML_KEM_768, keymgmt, prop, p11prov_mlkem768_keymgmt_functions);
+    ADD_ALGO_EXT(ML_KEM_1024, keymgmt, prop,
+                 p11prov_mlkem1024_keymgmt_functions);
     TERM_ALGO(keymgmt);
 
+#if SKEY_SUPPORT == 1
+    /* skeymgmt */
+    ADD_ALGO(AES, aes, skeymgmt, prop);
+    ADD_ALGO(GENERIC_SECRET, generic_secret, skeymgmt, prop);
+    TERM_ALGO(skeymgmt);
+#endif
+
     return CKR_OK;
 }
 
 static const char *p11prov_block_ops_names[OSSL_OP__HIGHEST + 1] = {
     [OSSL_OP_DIGEST] = "digest",
+#if SKEY_SUPPORT == 1
     [OSSL_OP_CIPHER] = "cipher",
+#endif
     [OSSL_OP_MAC] = "mac",
     [OSSL_OP_KDF] = "kdf",
     [OSSL_OP_RAND] = "rand",
@@ -1212,6 +1549,9 @@ static const char *p11prov_block_ops_names[OSSL_OP__HIGHEST + 1] = {
     [OSSL_OP_ENCODER] = "encoder",
     [OSSL_OP_DECODER] = "decoder",
     [OSSL_OP_STORE] = "store",
+#if SKEY_SUPPORT == 1
+    [OSSL_OP_SKEYMGMT] = "skeymgmt",
+#endif
 };
 
 static const OSSL_ALGORITHM *
@@ -1253,6 +1593,9 @@ p11prov_query_operation(void *provctx, int operation_id, int *no_cache)
     case OSSL_OP_ASYM_CIPHER:
         *no_cache = ctx->status == P11PROV_UNINITIALIZED ? 1 : 0;
         return ctx->op_asym_cipher;
+    case OSSL_OP_KEM:
+        *no_cache = ctx->status == P11PROV_UNINITIALIZED ? 1 : 0;
+        return ctx->op_kem;
     case OSSL_OP_ENCODER:
         *no_cache = 0;
         return ctx->op_encoder;
@@ -1262,6 +1605,14 @@ p11prov_query_operation(void *provctx, int operation_id, int *no_cache)
     case OSSL_OP_STORE:
         *no_cache = 0;
         return ctx->op_store;
+#if SKEY_SUPPORT == 1
+    case OSSL_OP_CIPHER:
+        *no_cache = 0;
+        return ctx->op_cipher;
+    case OSSL_OP_SKEYMGMT:
+        *no_cache = 0;
+        return ctx->op_skeymgmt;
+#endif
     }
     *no_cache = 0;
     return NULL;
@@ -1275,6 +1626,9 @@ static int p11prov_get_capabilities(void *provctx, const char *capability,
     if (OPENSSL_strcasecmp(capability, "TLS-GROUP") == 0) {
         ret = tls_group_capabilities(cb, arg);
     }
+    if (OPENSSL_strcasecmp(capability, "TLS-SIGALG") == 0) {
+        ret = tls_sigalg_capabilities(cb, arg);
+    }
 
     return ret;
 }
@@ -1464,6 +1818,9 @@ int OSSL_provider_init(const OSSL_CORE_HANDLE *handle, const OSSL_DISPATCH *in,
     }
     ctx->handle = handle;
 
+    P11PROV_debug("Starting provider %s %d.%d", PACKAGE_NAME, PACKAGE_MAJOR,
+                  PACKAGE_MINOR);
+
     ret = pthread_rwlock_init(&ctx->quirk_lock, NULL);
     if (ret != 0) {
         ret = errno;
