@@ -4,6 +4,15 @@
 
 source "${TESTSSRCDIR}/helpers.sh"
 
+_KEYID_COUNTER=0
+get_next_keyid() {
+    local id_val=${_KEYID_COUNTER}
+    ((_KEYID_COUNTER+=1))
+
+    KEYID=$(printf "%04x" ${id_val})
+    URIKEYID=$(printf "%%%02x%%%02x" $((id_val / 256)) $((id_val % 256)))
+}
+
 if [ $# -ne 1 ]; then
     echo "Usage setup.sh <tokentype>"
     exit 1
@@ -14,11 +23,14 @@ TOKENTYPE=$1
 # defaults -- overridden below or in the per-token setup
 SUPPORT_ED25519=1
 SUPPORT_ED448=1
+SUPPORT_X25519=1
+SUPPORT_X448=1
 SUPPORT_RSA_PKCS1_ENCRYPTION=1
 SUPPORT_RSA_KEYGEN_PUBLIC_EXPONENT=1
 SUPPORT_TLSFUZZER=1
 SUPPORT_ALLOWED_MECHANISMS=0
 SUPPORT_SYMMETRIC=1
+SUPPORT_BLOCK_MODES="CBC OFB CFB CFB1 CFB8 CTR ECB"
 
 # Ed448 requires OpenSC 0.26.0
 OPENSC_VERSION=$(opensc-tool -i | grep OpenSC | sed -e "s/OpenSC 0\.\([0-9]*\).*/\1/")
@@ -31,6 +43,9 @@ if [[ "${PKCS11_PROVIDER_FORCE_FIPS_MODE}" = "1" || "$(cat /proc/sys/crypto/fips
     # We can not use Edwards curves in FIPS mode
     SUPPORT_ED25519=0
     SUPPORT_ED448=0
+    # We can not use Montgomery curves in FIPS mode
+    SUPPORT_X25519=0
+    SUPPORT_X448=0
 
     # The FIPS does not allow the RSA-PKCS1.5 encryption
     SUPPORT_RSA_PKCS1_ENCRYPTION=0
@@ -53,9 +68,9 @@ if [[ "${PKCS11_PROVIDER_FORCE_FIPS_MODE}" = "1" || "$(cat /proc/sys/crypto/fips
     export NSS_FIPS=1
 
     # NSS softokn requires stronger PIN in FIPS mode
-    PINVALUE="fo0m4nchU"
+    PINVALUE="fo0m4nchU-p4sSw0rd-iS-Str0ng1234"
 else
-    PINVALUE="12345678"
+    PINVALUE="0123456789ABCDEFFEDCBA9876543210"
 fi
 
 # Check if openssl supports skey
@@ -146,12 +161,11 @@ crt_selfsign() {
 }
 
 title LINE "Creating new Self Sign CA"
-KEYID='0000'
-URIKEYID="%00%00"
+get_next_keyid
 CACRTN="caCert"
 ptool --keypairgen --key-type="RSA:2048" --id="${KEYID}" \
       --label="${CACRTN}" 2>&1
-crt_selfsign $CACRTN "Issuer" $KEYID
+crt_selfsign $CACRTN "Issuer" "${KEYID}"
 
 # convert the DER cert to PEM
 CACRT_PEM="${TMPPDIR}/${CACRTN}.pem"
@@ -207,13 +221,12 @@ ca_sign() {
 
 
 # generate RSA key pair and self-signed certificate
-KEYID='0001'
-URIKEYID="%00%01"
+get_next_keyid
 TSTCRTN="testCert"
 
 ptool --keypairgen --key-type="RSA:2048" --id="$KEYID" \
       --label="${TSTCRTN}" 2>&1
-ca_sign "${TSTCRTN}" "My Test Cert" $KEYID
+ca_sign "${TSTCRTN}" "My Test Cert" "$KEYID"
 
 BASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
 BASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
@@ -232,13 +245,12 @@ echo "${CRTURI}"
 echo ""
 
 # generate ECC key pair
-KEYID='0002'
-URIKEYID="%00%02"
+get_next_keyid
 ECCRTN="ecCert"
 
 ptool --keypairgen --key-type="EC:secp256r1" --id="$KEYID" \
       --label="${ECCRTN}" 2>&1
-ca_sign $ECCRTN "My EC Cert" $KEYID
+ca_sign $ECCRTN "My EC Cert" "$KEYID"
 
 ECBASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
 ECBASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
@@ -247,13 +259,12 @@ ECPUBURI="pkcs11:type=public;id=${URIKEYID}"
 ECPRIURI="pkcs11:type=private;id=${URIKEYID}"
 ECCRTURI="pkcs11:type=cert;object=${ECCRTN}"
 
-KEYID='0003'
-URIKEYID="%00%03"
+get_next_keyid
 ECPEERCRTN="ecPeerCert"
 
 ptool --keypairgen --key-type="EC:secp256r1" --id="$KEYID" \
       --label="$ECPEERCRTN" 2>&1
-crt_selfsign $ECPEERCRTN "My Peer EC Cert" $KEYID
+crt_selfsign $ECPEERCRTN "My Peer EC Cert" "$KEYID"
 
 ECPEERBASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
 ECPEERBASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
@@ -281,13 +292,12 @@ echo ""
 ## Softtokn does not support edwards curves yet
 if [ "${SUPPORT_ED25519}" -eq 1 ]; then
     # generate ED25519
-    KEYID='0004'
-    URIKEYID="%00%04"
+    get_next_keyid
     EDCRTN="edCert"
 
     ptool --keypairgen --key-type="EC:edwards25519" --id="$KEYID" \
     	  --label="${EDCRTN}" 2>&1
-    ca_sign $EDCRTN "My ED25519 Cert" $KEYID
+    ca_sign $EDCRTN "My ED25519 Cert" "$KEYID"
 
     EDBASEURIWITHPINVALUE="pkcs11:id=${URIKEYID};pin-value=${PINVALUE}"
     EDBASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID};pin-source=file:${PINFILE}"
@@ -307,13 +317,12 @@ fi
 
 if [ "${SUPPORT_ED448}" -eq 1 ]; then
     # generate ED448
-    KEYID='0009'
-    URIKEYID="%00%09"
+    get_next_keyid
     ED2CRTN="ed2Cert"
 
     ptool --keypairgen --key-type="EC:Ed448" --id="$KEYID" \
           --label="${ED2CRTN}" 2>&1
-    ca_sign $ED2CRTN "My ED448 Cert" $KEYID
+    ca_sign $ED2CRTN "My ED448 Cert" "$KEYID"
 
     ED2BASEURIWITHPINVALUE="pkcs11:id=${URIKEYID};pin-value=${PINVALUE}"
     ED2BASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID};pin-source=file:${PINFILE}"
@@ -331,15 +340,82 @@ if [ "${SUPPORT_ED448}" -eq 1 ]; then
     echo "${ED2CRTURI}"
 fi
 
+if [ "${SUPPORT_X25519}" -eq 1 ]; then
+    # generate X25519
+    get_next_keyid
+
+    ptool --keypairgen --key-type="EC:X25519" --id="$KEYID" \
+    	  --label="x25519 key" 2>&1
+
+    X25519BASEURIWITHPINVALUE="pkcs11:id=${URIKEYID};pin-value=${PINVALUE}"
+    X25519BASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID};pin-source=file:${PINFILE}"
+    X25519BASEURI="pkcs11:id=${URIKEYID}"
+    X25519PUBURI="pkcs11:type=public;id=${URIKEYID}"
+    X25519PRIURI="pkcs11:type=private;id=${URIKEYID}"
+
+    title LINE "X25519 PKCS11 URIS"
+    echo "${X25519BASEURIWITHPINVALUE}"
+    echo "${X25519BASEURIWITHPINSOURCE}"
+    echo "${X25519BASEURI}"
+    echo "${X25519PUBURI}"
+    echo "${X25519PRIURI}"
+
+    # And a Peer Key for Key Exchange (ECDH) tests
+    get_next_keyid
+
+    ptool --keypairgen --key-type="EC:X25519" --id="$KEYID" \
+    	  --label="x25519 key" 2>&1
+
+    X25519PEERPUBURI="pkcs11:type=public;id=${URIKEYID}"
+    X25519PEERPRIURI="pkcs11:type=private;id=${URIKEYID}"
+
+    title LINE "X25519 Peer PKCS11 URIS"
+    echo "${X25519PEERPUBURI}"
+    echo "${X25519PEERPRIURI}"
+fi
+
+if [ "${SUPPORT_X448}" -eq 1 ]; then
+    # generate X448
+    get_next_keyid
+
+    ptool --keypairgen --key-type="EC:X448" --id="$KEYID" \
+    	  --label="x448 key" 2>&1
+
+    X448BASEURIWITHPINVALUE="pkcs11:id=${URIKEYID};pin-value=${PINVALUE}"
+    X448BASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID};pin-source=file:${PINFILE}"
+    X448BASEURI="pkcs11:id=${URIKEYID}"
+    X448PUBURI="pkcs11:type=public;id=${URIKEYID}"
+    X448PRIURI="pkcs11:type=private;id=${URIKEYID}"
+
+    title LINE "X448 PKCS11 URIS"
+    echo "${X448BASEURIWITHPINVALUE}"
+    echo "${X448BASEURIWITHPINSOURCE}"
+    echo "${X448BASEURI}"
+    echo "${X448PUBURI}"
+    echo "${X448PRIURI}"
+
+    # And a Peer Key for Key Exchange (ECDH) tests
+    get_next_keyid
+
+    ptool --keypairgen --key-type="EC:X448" --id="$KEYID" \
+    	  --label="x448 key" 2>&1
+
+    X448PEERPUBURI="pkcs11:type=public;id=${URIKEYID}"
+    X448PEERPRIURI="pkcs11:type=private;id=${URIKEYID}"
+
+    title LINE "X448 Peer PKCS11 URIS"
+    echo "${X448PEERPUBURI}"
+    echo "${X448PEERPRIURI}"
+fi
+
 title PARA "generate RSA key pair, self-signed certificate, remove public key"
-KEYID='0005'
-URIKEYID="%00%05"
+get_next_keyid
 TSTCRTN="testCert2"
 
 ptool --keypairgen --key-type="RSA:2048" --id="$KEYID" \
       --label="${TSTCRTN}" 2>&1
-ca_sign $TSTCRTN "My Test Cert 2" $KEYID
-ptool --delete-object --type pubkey --id 0005 2>&1
+ca_sign $TSTCRTN "My Test Cert 2" "$KEYID"
+ptool --delete-object --type pubkey --id "$KEYID" 2>&1
 
 BASE2URIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
 BASE2URIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=${PINFILE}"
@@ -356,14 +432,13 @@ echo "${CRT2URI}"
 echo ""
 
 title PARA "generate EC key pair, self-signed certificate, remove public key"
-KEYID='0006'
-URIKEYID="%00%06"
+get_next_keyid
 TSTCRTN="ecCert2"
 
 ptool --keypairgen --key-type="EC:secp384r1" --id="$KEYID" \
       --label="${TSTCRTN}" 2>&1
-ca_sign $TSTCRTN "My EC Cert 2" $KEYID
-ptool --delete-object --type pubkey --id 0006 2>&1
+ca_sign $TSTCRTN "My EC Cert 2" "$KEYID"
+ptool --delete-object --type pubkey --id "$KEYID" 2>&1
 
 ECBASE2URIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
 ECBASE2URIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file${PINFILE}"
@@ -385,8 +460,7 @@ elif [ "${TOKENTYPE}" == "softokn" ]; then
     title PARA "explicit EC unsupported with softokn"
 else
     title PARA "generate explicit EC key pair"
-    KEYID='0007'
-    URIKEYID="%00%07"
+    get_next_keyid
     ECXCRTN="ecExplicitCert"
 
     ptool --write-object="${TESTSSRCDIR}/explicit_ec.key.der" --type=privkey \
@@ -408,13 +482,12 @@ else
 fi
 
 title PARA "generate EC key pair with ALWAYS AUTHENTICATE flag, self-signed certificate"
-KEYID='0008'
-URIKEYID="%00%08"
+get_next_keyid
 TSTCRTN="ecCert3"
 
 ptool --keypairgen --key-type="EC:secp521r1" --id="$KEYID" \
       --label="${TSTCRTN}" --always-auth 2>&1
-ca_sign $TSTCRTN "My EC Cert 3" $KEYID
+ca_sign $TSTCRTN "My EC Cert 3" "$KEYID"
 
 ECBASE3URIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
 ECBASE3URIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
@@ -434,8 +507,7 @@ echo ""
 
 if [ "${SUPPORT_ALLOWED_MECHANISMS}" -eq 1 ]; then
     # generate unrestricted RSA-PSS key pair and RSA-PSS certificate
-    KEYID='0010'
-    URIKEYID="%00%10"
+    get_next_keyid
     TSTCRTN="testRsaPssCert"
     MECHS="RSA-PKCS-PSS"
     MECHS+=",SHA1-RSA-PKCS-PSS,SHA224-RSA-PKCS-PSS"
@@ -443,7 +515,7 @@ if [ "${SUPPORT_ALLOWED_MECHANISMS}" -eq 1 ]; then
 
     ptool --keypairgen --key-type="RSA:2048" --id="$KEYID" \
           --label="${TSTCRTN}" --allowed-mechanisms "$MECHS" 2>&1
-    ca_sign "${TSTCRTN}" "My RsaPss Cert" $KEYID "PSS"
+    ca_sign "${TSTCRTN}" "My RsaPss Cert" "$KEYID" "PSS"
 
     RSAPSSBASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
     RSAPSSBASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
@@ -463,14 +535,13 @@ if [ "${SUPPORT_ALLOWED_MECHANISMS}" -eq 1 ]; then
 
     # generate RSA-PSS (3k) key pair restricted to SHA256 digests
     # and RSA-PSS certificate
-    KEYID='0011'
-    URIKEYID="%00%11"
+    get_next_keyid
     TSTCRTN="testRsaPss2Cert"
     MECHS="SHA256-RSA-PKCS-PSS"
 
     ptool --keypairgen --key-type="RSA:3092" --id="$KEYID" \
           --label="${TSTCRTN}" --allowed-mechanisms "$MECHS" 2>&1
-    ca_sign "${TSTCRTN}" "My RsaPss2 Cert" $KEYID "PSS-SHA256"
+    ca_sign "${TSTCRTN}" "My RsaPss2 Cert" "$KEYID" "PSS-SHA256"
 
     RSAPSS2BASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
     RSAPSS2BASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
@@ -491,8 +562,7 @@ fi
 
 if [ "$SUPPORT_ML_DSA" -eq 1 ]; then
     title PARA "generate ML-DSA Key pair"
-    KEYID='0012'
-    URIKEYID="%00%12"
+    get_next_keyid
     TSTCRTN="mlDsa"
 
     # not supported by the pkcs11-tool yet. Do it for now with OpenSSL CLI
@@ -509,7 +579,7 @@ if [ "$SUPPORT_ML_DSA" -eq 1 ]; then
             -algorithm ML-DSA-44
             -pkeyopt "pkcs11_uri:pkcs11:object=${TSTCRTN};id=${URIKEYID}"'
     OPENSSL_CONF=${ORIG_OPENSSL_CONF}
-    ca_sign $TSTCRTN "My ML-DSA Cert" $KEYID
+    ca_sign $TSTCRTN "My ML-DSA Cert" "$KEYID"
 
     MLDSABASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
     MLDSABASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
@@ -526,6 +596,37 @@ if [ "$SUPPORT_ML_DSA" -eq 1 ]; then
     echo ""
 fi
 
+if [ "$SUPPORT_ML_KEM" -eq 1 ]; then
+    title PARA "generate ML-KEM Key pair"
+    get_next_keyid
+    TSTCRTN="mlKem"
+
+    # not supported by the pkcs11-tool yet. Do it for now with OpenSSL CLI
+    ORIG_OPENSSL_CONF=${OPENSSL_CONF}
+    # We need to configure pkcs11 to allow emitting PEM URIs so that the
+    # genpkey command does not fail on trying to emit the private key PEM file.
+    sed -e "s/#pkcs11-module-encode-provider-uri-to-pem/pkcs11-module-encode-provider-uri-to-pem = true/" \
+        "${OPENSSL_CONF}" > "${OPENSSL_CONF}.mlkem_pem_uri"
+    OPENSSL_CONF=${OPENSSL_CONF}.mlkem_pem_uri
+    ossl '
+    genpkey -propquery "provider=pkcs11"
+            -algorithm ML-KEM-512
+            -pkeyopt "pkcs11_uri:pkcs11:object=${TSTCRTN};id=${URIKEYID}"'
+    OPENSSL_CONF=${ORIG_OPENSSL_CONF}
+
+    MLKEMBASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
+    MLKEMBASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
+    MLKEMBASEURI="pkcs11:id=${URIKEYID}"
+    MLKEMPUBURI="pkcs11:type=public;id=${URIKEYID}"
+    MLKEMPRIURI="pkcs11:type=private;id=${URIKEYID}"
+
+    title LINE "ML-KEM PKCS11 URIS"
+    echo "${MLKEMBASEURI}"
+    echo "${MLKEMPUBURI}"
+    echo "${MLKEMPRIURI}"
+    echo ""
+fi
+
 
 title PARA "Show contents of ${TOKENTYPE} token"
 echo " ----------------------------------------------------------------------------------------------------"
@@ -547,13 +648,17 @@ export TESTBLDDIR="${TESTBLDDIR}"
 
 export SUPPORT_ED25519="${SUPPORT_ED25519}"
 export SUPPORT_ED448="${SUPPORT_ED448}"
+export SUPPORT_X25519="${SUPPORT_X25519}"
+export SUPPORT_X448="${SUPPORT_X448}"
 export SUPPORT_ML_DSA="${SUPPORT_ML_DSA}"
+export SUPPORT_ML_KEM="${SUPPORT_ML_KEM}"
 export SUPPORT_RSA_PKCS1_ENCRYPTION="${SUPPORT_RSA_PKCS1_ENCRYPTION}"
 export SUPPORT_RSA_KEYGEN_PUBLIC_EXPONENT="${SUPPORT_RSA_KEYGEN_PUBLIC_EXPONENT}"
 export SUPPORT_TLSFUZZER="${SUPPORT_TLSFUZZER}"
 export SUPPORT_ALLOWED_MECHANISMS="${SUPPORT_ALLOWED_MECHANISMS}"
 export SUPPORT_SKEY="${SUPPORT_SKEY}"
 export SUPPORT_SYMMETRIC="${SUPPORT_SYMMETRIC}"
+export SUPPORT_BLOCK_MODES="${SUPPORT_BLOCK_MODES}"
 
 export TESTPORT="${TESTPORT}"
 
@@ -636,6 +741,34 @@ export ED2CRTURI="${ED2CRTURI}"
 DBGSCRIPT
 fi
 
+if [ -n "${X25519BASEURI}" ]; then
+    cat >> "${TMPPDIR}/testvars" <<DBGSCRIPT
+
+export X25519BASEURIWITHPINVALUE="${X25519BASEURIWITHPINVALUE}"
+export X25519BASEURIWITHPINSOURCE="${X25519BASEURIWITHPINSOURCE}"
+export X25519BASEURI="${X25519BASEURI}"
+export X25519PUBURI="${X25519PUBURI}"
+export X25519PRIURI="${X25519PRIURI}"
+
+export X25519PEERPUBURI="${X25519PEERPUBURI}"
+export X25519PEERPRIURI="${X25519PEERPRIURI}"
+DBGSCRIPT
+fi
+
+if [ -n "${X448BASEURI}" ]; then
+    cat >> "${TMPPDIR}/testvars" <<DBGSCRIPT
+
+export X448BASEURIWITHPINVALUE="${X448BASEURIWITHPINVALUE}"
+export X448BASEURIWITHPINSOURCE="${X448BASEURIWITHPINSOURCE}"
+export X448BASEURI="${X448BASEURI}"
+export X448PUBURI="${X448PUBURI}"
+export X448PRIURI="${X448PRIURI}"
+
+export X448PEERPUBURI="${X448PEERPUBURI}"
+export X448PEERPRIURI="${X448PEERPRIURI}"
+DBGSCRIPT
+fi
+
 if [ -n "${ECXBASEURI}" ]; then
     cat >> "${TMPPDIR}/testvars" <<DBGSCRIPT
 
@@ -678,6 +811,17 @@ export MLDSACRTURI="${MLDSACRTURI}"
 DBGSCRIPT
 fi
 
+if [ -n "${MLKEMBASEURI}" ]; then
+    cat >> "${TMPPDIR}/testvars" <<DBGSCRIPT
+
+export MLKEMBASEURIWITHPINVALUE="${MLKEMBASEURIWITHPINVALUE}"
+export MLKEMBASEURIWITHPINSOURCE="${MLKEMBASEURIWITHPINSOURCE}"
+export MLKEMBASEURI="${MLKEMBASEURI}"
+export MLKEMPUBURI="${MLKEMPUBURI}"
+export MLKEMPRIURI="${MLKEMPRIURI}"
+DBGSCRIPT
+fi
+
 cat >> "${TMPPDIR}/testvars" <<DBGSCRIPT
 
 # for listing the separate pkcs11 calls
@@ -700,4 +844,3 @@ fi
 gen_unsetvars
 
 title ENDSECTION
-
