@@ -33,4 +33,11 @@ export TESTPORT="30000"
 export SUPPORT_ED25519=0
 export SUPPORT_ED448=0
 
+# Montgomery curves are not supported in NSS yet
+export SUPPORT_X25519=0
+export SUPPORT_X448=0
+
 export SUPPORT_ML_DSA=0
+export SUPPORT_ML_KEM=0
+
+export SUPPORT_BLOCK_MODES="CBC CTR ECB"
